@@ -9,6 +9,7 @@ from types import TracebackType
 
 from sql_tstring.parser import (
     Clause,
+    ClauseGroup,
     Element,
     Expression,
     ExpressionGroup,
@@ -25,7 +26,7 @@ from sql_tstring.parser import (
 from sql_tstring.t import t, Template as TTemplate
 
 try:
-    from string.templatelib import Template  # type: ignore[import-untyped]
+    from string.templatelib import Template
 except ImportError:
 
     class Template:  # type: ignore[no-redef]
@@ -87,7 +88,7 @@ def sql_context(
 def sql(
     query_or_template: str | Template | TTemplate, values: dict[str, typing.Any] | None = None
 ) -> tuple[str, list]:
-    template: Template
+    template: Template | TTemplate
     if isinstance(query_or_template, (Template, TTemplate)) and values is None:
         template = query_or_template
     elif isinstance(query_or_template, str) and values is not None:
@@ -102,10 +103,10 @@ def sql(
     for raw_parsed_query in parsed_queries:
         parsed_query = deepcopy(raw_parsed_query)
         new_values = _replace_placeholders(parsed_query, 0)
-        result_str += _print_node(parsed_query, [None] * len(result_values), ctx.dialect)
+        result_str += " " + _print_node(parsed_query, [None] * len(result_values), ctx.dialect)
         result_values.extend(new_values)
 
-    return result_str, result_values
+    return result_str.strip(), result_values
 
 
 class _ContextManager:
@@ -131,12 +132,15 @@ def _check_valid(
     *,
     case_sensitive: set[str] | None = None,
     case_insensitive: set[str] | None = None,
+    value_type: type = str,
 ) -> None:
     if case_sensitive is None:
         case_sensitive = set()
     if case_insensitive is None:
         case_insensitive = set()
-    if not isinstance(value, str) or (
+    if not isinstance(value, value_type):
+        raise ValueError(f"{value} is not valid, must be {value_type}")
+    if isinstance(value, str) and (
         value not in case_sensitive and value.lower() not in case_insensitive
     ):
         raise ValueError(
@@ -154,7 +158,12 @@ def _print_node(
 
     match node:
         case Statement():
-            result = " ".join(_print_node(clause, placeholders, dialect) for clause in node.clauses)
+            addition = " ".join(
+                _print_node(clause, placeholders, dialect) for clause in node.clauses
+            )
+            result = f"{node.separator} {addition}"
+        case ClauseGroup():
+            result = f"({" ".join(_print_node(clause, placeholders, dialect) for clause in node.clauses)})"  # noqa: E501
         case Clause() | ExpressionGroup():
             result = ""
 
@@ -274,36 +283,40 @@ def _replace_placeholder(
                     )
                     result.append(value)
     else:
-        if clause is not None and clause.text.lower() == "order by":
-            _check_valid(
-                value,
-                case_sensitive=ctx.columns,
-                case_insensitive={"asc", "ascending", "desc", "descending"},
-            )
-            new_node = Part(text=typing.cast(str, value), parent=node.parent)
-        elif placeholder_type == PlaceholderType.COLUMN:
-            _check_valid(value, case_sensitive=ctx.columns)
-            new_node = Part(text=typing.cast(str, value), parent=node.parent)
-        elif placeholder_type == PlaceholderType.TABLE:
-            _check_valid(value, case_sensitive=ctx.tables)
-            new_node = Part(text=typing.cast(str, value), parent=node.parent)
-        elif placeholder_type == PlaceholderType.LOCK:
-            _check_valid(value, case_insensitive={"", "nowait", "skip locked"})
-            new_node = Part(text=typing.cast(str, value), parent=node.parent)
-        else:
-            if (
-                value is RewritingValue.IS_NULL or value is RewritingValue.IS_NOT_NULL
-            ) and placeholder_type == PlaceholderType.VARIABLE_CONDITION:
-                for part in node.parent.parts:
-                    if isinstance(part, Operator):
-                        if value is RewritingValue.IS_NULL:
-                            part.text = "IS"
-                        else:
-                            part.text = "IS NOT"
-                new_node = Part(text="NULL", parent=node.parent)
-            else:
-                new_node = node
-                result.append(value)
+        match placeholder_type:
+            case PlaceholderType.COLUMN:
+                _check_valid(value, case_sensitive=ctx.columns)
+                new_node = Part(text=typing.cast(str, value), parent=node.parent)
+            case PlaceholderType.FRAME:
+                _check_valid(value, value_type=int)
+                new_node = Part(text=str(value), parent=node.parent)
+            case PlaceholderType.LOCK:
+                _check_valid(value, case_insensitive={"", "nowait", "skip locked"})
+                new_node = Part(text=typing.cast(str, value), parent=node.parent)
+            case PlaceholderType.SORT:
+                _check_valid(
+                    value,
+                    case_sensitive=ctx.columns,
+                    case_insensitive={"asc", "ascending", "desc", "descending"},
+                )
+                new_node = Part(text=typing.cast(str, value), parent=node.parent)
+            case PlaceholderType.TABLE:
+                _check_valid(value, case_sensitive=ctx.tables)
+                new_node = Part(text=typing.cast(str, value), parent=node.parent)
+            case _:
+                if (
+                    value is RewritingValue.IS_NULL or value is RewritingValue.IS_NOT_NULL
+                ) and placeholder_type == PlaceholderType.VARIABLE_CONDITION:
+                    for part in node.parent.parts:
+                        if isinstance(part, Operator):
+                            if value is RewritingValue.IS_NULL:
+                                part.text = "IS"
+                            else:
+                                part.text = "IS NOT"
+                    new_node = Part(text="NULL", parent=node.parent)
+                else:
+                    new_node = node
+                    result.append(value)
 
         if isinstance(node.parent, (Expression, ExpressionGroup, Function, Group)):
             node.parent.parts[index] = new_node
