@@ -5,12 +5,18 @@ from dataclasses import dataclass, field
 from enum import auto, Enum, unique
 from typing import cast
 
-from sql_tstring.t import Template as TTemplate
+from sql_tstring.t import Interpolation as TInterpolation, Template as TTemplate
 
 try:
-    from string.templatelib import Interpolation, Template  # type: ignore[import-untyped]
+    from string.templatelib import Interpolation, Template
 except ImportError:
-    from sql_tstring.t import Interpolation, Template
+
+    class Interpolation:  # type: ignore[no-redef]
+        pass
+
+    class Template:  # type: ignore[no-redef]
+        pass
+
 
 SPLIT_RE = re.compile(r"([^\s'(]+\(|\(|'+|[ ',;)\n\t])")
 
@@ -19,7 +25,9 @@ SPLIT_RE = re.compile(r"([^\s'(]+\(|\(|'+|[ ',;)\n\t])")
 class PlaceholderType(Enum):
     COLUMN = auto()
     DISALLOWED = auto()
+    FRAME = auto()
     LOCK = auto()
+    SORT = auto()
     TABLE = auto()
     VARIABLE = auto()
     VARIABLE_CONDITION = auto()
@@ -117,6 +125,13 @@ CLAUSES: ClauseDictionary = {
         ),
     },
     "order": {
+        "by": {
+            "": ClauseProperties(
+                allow_empty=False, placeholder_type=PlaceholderType.SORT, separators={","}
+            )
+        },
+    },
+    "partition": {
         "by": {
             "": ClauseProperties(
                 allow_empty=False, placeholder_type=PlaceholderType.COLUMN, separators={","}
@@ -152,6 +167,11 @@ CLAUSES: ClauseDictionary = {
             allow_empty=False, placeholder_type=PlaceholderType.TABLE, separators=set()
         )
     },
+    "groups": {
+        "": ClauseProperties(
+            allow_empty=False, placeholder_type=PlaceholderType.FRAME, separators=set()
+        )
+    },
     "having": {
         "": ClauseProperties(
             allow_empty=False,
@@ -172,11 +192,21 @@ CLAUSES: ClauseDictionary = {
             allow_empty=False, placeholder_type=PlaceholderType.VARIABLE, separators=set()
         )
     },
+    "range": {
+        "": ClauseProperties(
+            allow_empty=False, placeholder_type=PlaceholderType.FRAME, separators=set()
+        )
+    },
     "returning": {
         "": ClauseProperties(
             allow_empty=False, placeholder_type=PlaceholderType.DISALLOWED, separators={","}
         )
     },
+    "rows": {
+        "": ClauseProperties(
+            allow_empty=False, placeholder_type=PlaceholderType.FRAME, separators=set()
+        )
+    },
     "select": {
         "": ClauseProperties(
             allow_empty=False, placeholder_type=PlaceholderType.COLUMN, separators={","}
@@ -244,13 +274,14 @@ OPERATORS: dict[str, dict] = {
 
 @dataclass
 class Statement:
-    clauses: list[Clause] = field(default_factory=list)
+    clauses: list[Clause | ClauseGroup] = field(default_factory=list)
     parent: ExpressionGroup | Function | Group | None = None
+    separator: str = ""
 
 
 @dataclass
 class Clause:
-    parent: Statement
+    parent: ClauseGroup | Statement
     properties: ClauseProperties
     text: str
     expressions: list[Expression] = field(init=False)
@@ -260,6 +291,12 @@ class Clause:
         self.expressions = [Expression(self)]
 
 
+@dataclass
+class ClauseGroup:
+    parent: Statement
+    clauses: list[Clause] = field(default_factory=list)
+
+
 @dataclass
 class Expression:
     parent: Clause | ExpressionGroup
@@ -321,21 +358,28 @@ class Operator:
 
 
 type ParentNode = Clause | Expression | ExpressionGroup | Function | Group
-type Node = ParentNode | Literal | Statement
+type Node = ParentNode | ClauseGroup | Literal | Statement
 type Element = Node | Operator | Part | Placeholder
 
 
-def parse(template: Template) -> list[Statement]:
+def parse(template: Template | TTemplate) -> list[Statement]:
     statements = [Statement()]
     current_node: Node = statements[0]
     _parse_template(template, current_node, statements)
     return statements
 
 
-def _parse_template(template: Template, current_node: Node, statements: list[Statement]) -> None:
+def _parse_template(
+    template: Template | TTemplate, current_node: Node, statements: list[Statement]
+) -> None:
     for item in template:
         match item:
-            case Interpolation(value, _, _, _):  # type: ignore[misc]
+            case Interpolation(value, _, _, _):
+                if isinstance(value, (Template, TTemplate)):
+                    _parse_template(value, current_node, statements)
+                else:
+                    _parse_placeholder(current_node, value)
+            case TInterpolation(value, _, _, _):
                 if isinstance(value, (Template, TTemplate)):
                     _parse_template(value, current_node, statements)
                 else:
@@ -350,7 +394,7 @@ def _parse_placeholder(
 ) -> None:
     if isinstance(current_node, (Expression, Function, Group, Literal)):
         parent = current_node
-    elif isinstance(current_node, Statement):
+    elif isinstance(current_node, (Statement, ClauseGroup)):
         raise ValueError("Invalid syntax")
     else:  # Clause | ExpressionGroup
         parent = current_node.expressions[-1]
@@ -372,13 +416,17 @@ def _parse_string(
         consumed = 1
         if isinstance(current_node, Literal):
             if current_token == "'":
-                current_node = current_node.parent
+                current_node = _find_node(  # type: ignore[assignment]
+                    current_node.parent, (Clause, ExpressionGroup, Function, Group)
+                )
             else:
                 current_node.parts.append(Part(parent=current_node, text=raw_current_token))
         elif isinstance(current_node, (Function, Group)):
             if current_token == ")":
                 group_or_function = _find_node(current_node, (Function, Group))
-                current_node = group_or_function.parent
+                current_node = _find_node(  # type: ignore[assignment]
+                    group_or_function.parent, (Clause, ExpressionGroup, Function, Group)
+                )
             else:
                 current_node, consumed = _parse_token(
                     current_node, raw_current_token, current_token, tokens[index:], statements
@@ -386,7 +434,9 @@ def _parse_string(
         elif isinstance(current_node, ExpressionGroup):
             if current_token == ")":
                 group = _find_node(current_node, ExpressionGroup)
-                current_node = group.parent
+                current_node = _find_node(  # type: ignore[assignment]
+                    group.parent, (Clause, ExpressionGroup)
+                )
             else:
                 clause = _find_node(current_node, Clause)
                 if current_token in clause.properties.separators:
@@ -406,7 +456,19 @@ def _parse_string(
                 current_node, consumed = _parse_token(
                     current_node, raw_current_token, current_token, tokens[index:], statements
                 )
-        else:  # Expression | Statement
+        elif isinstance(current_node, Expression):
+            clause = _find_node(current_node, Clause)
+            if current_token in clause.properties.separators:
+                parent_group = cast(
+                    Clause | ExpressionGroup, _find_node(current_node, (Clause, ExpressionGroup))
+                )
+                current_node = Expression(parent=parent_group, separator=raw_current_token)
+                parent_group.expressions.append(current_node)
+            else:
+                current_node, consumed = _parse_token(
+                    current_node, raw_current_token, current_token, tokens[index:], statements
+                )
+        else:  # ClauseGroup | Statement
             current_node, consumed = _parse_token(
                 current_node, raw_current_token, current_token, tokens[index:], statements
             )
@@ -417,7 +479,7 @@ def _parse_string(
 
 
 def _parse_token(
-    current_node: ParentNode | Statement,
+    current_node: ParentNode | ClauseGroup | Statement,
     raw_current_token: str,
     current_token: str,
     tokens: list[str],
@@ -427,8 +489,17 @@ def _parse_token(
         return _parse_clause(current_node, tokens)
     elif current_token == ";":
         statements.append(Statement())
+        statements[-1].separator = ";"
         return statements[-1], 1
-    elif not isinstance(current_node, Statement):
+    elif current_token == "union":
+        statements.append(Statement())
+        statements[-1].separator = raw_current_token
+        consumed = 1
+        if tokens[1].lower() == "all":
+            statements[-1].separator += f" {tokens[1]}"
+            consumed = 2
+        return statements[-1], consumed
+    elif not isinstance(current_node, (ClauseGroup, Statement)):
         if current_token in OPERATORS:
             return _parse_operator(current_node, tokens)
         elif current_token == "'":
@@ -439,17 +510,21 @@ def _parse_token(
             return _parse_function(current_node, raw_current_token[:-1])
         elif current_token == ")":
             current_node = _find_node(  # type: ignore[assignment]
-                current_node, (ExpressionGroup, Function, Group)
+                current_node, (ExpressionGroup, Function, Group, ClauseGroup)
             )
             return current_node.parent, 1
         else:
             return _parse_part(current_node, raw_current_token)
+    elif isinstance(current_node, Statement) and current_token == "(":
+        statement_group = ClauseGroup(current_node)
+        current_node.clauses.append(statement_group)
+        return statement_group, 1
     else:
         raise ValueError("Invalid syntax")
 
 
 def _parse_clause(
-    current_node: ParentNode | Statement,
+    current_node: ParentNode | ClauseGroup | Statement,
     tokens: list[str],
 ) -> tuple[Clause, int]:
     index = 0
@@ -468,16 +543,16 @@ def _parse_clause(
         statement = Statement(parent=current_node)
         current_node.expressions[-1].parts.append(statement)
         current_node = statement
-    else:  # Clause | Expression | Statement
-        current_node = _find_node(current_node, Statement)
+    else:  # Clause | Expression | Statement | ClauseGroup
+        current_node = _find_node(current_node, (Statement, ClauseGroup))  # type: ignore[assignment] # noqa: E501
 
     clause_properties = cast(ClauseProperties, clause_entry[""])
     clause = Clause(
-        parent=current_node,
+        parent=current_node,  # type: ignore[arg-type]
         properties=clause_properties,
         text=text,
     )
-    current_node.clauses.append(clause)
+    current_node.clauses.append(clause)  # type: ignore[union-attr]
     return clause, index
 
 
@@ -495,7 +570,7 @@ def _parse_operator[T: ParentNode](
     if isinstance(current_node, (Expression, Function, Group)):
         parent = current_node
     else:  # Clause | ExpressionGroup
-        parent = current_node.expressions[-1]
+        parent = current_node.expressions[-1]  # type: ignore[assignment]
     parent.parts.append(Operator(parent=parent, text=text))
     return current_node, index
 
@@ -550,7 +625,7 @@ def _parse_part[T: ParentNode](
     if isinstance(current_node, (Expression, Function, Group)):
         parent = current_node
     else:  # Clause | ExpressionGroup
-        parent = current_node.expressions[-1]
+        parent = current_node.expressions[-1]  # type: ignore[assignment]
     parent.parts.append(Part(parent=parent, text=text))
     return current_node, 1
 
